--
-- Drivable
--
-- @author  Stefan Geiger
-- @date  30/11/08
--
-- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.

source("dataS/scripts/vehicles/specializations/events/SetCruiseControlStateEvent.lua");
source("dataS/scripts/vehicles/specializations/events/SetCruiseControlSpeedEvent.lua");
source("dataS/scripts/vehicles/specializations/events/DrivableToggleLowerAllEvent.lua");

---Specialization class for Drivables
-- @category Specializations
customDrivable = {};

customDrivable.CRUISECONTROL_STATE_OFF = 0;
customDrivable.CRUISECONTROL_STATE_ACTIVE = 1;
customDrivable.CRUISECONTROL_STATE_FULL = 2;

---Checks if all prerequisite specializations are loaded
-- @param table specializations specializations
-- @return boolean hasPrerequisite true if all prerequisite specializations are loaded
-- @includeCode
function customDrivable.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Motorized, specializations) and
           SpecializationUtil.hasSpecialization(customSteerable, specializations) and
           SpecializationUtil.hasSpecialization(customLights, specializations) and
           SpecializationUtil.hasSpecialization(AttacherJoints, specializations);
end

---Called before loading
-- @param table savegame savegame
-- @includeCode
function customDrivable:preLoad(savegame)
    self.setOperatingTime = Utils.overwrittenFunction(self.setOperatingTime, customDrivable.setOperatingTime);
    self.getIsOperating = Utils.overwrittenFunction(self.getIsOperating, customDrivable.getIsOperating);
    self.getRpmLimit = Utils.overwrittenFunction(self.getRpmLimit, customDrivable.getRpmLimit);
    self.getDirtMultiplier = Utils.overwrittenFunction(self.getDirtMultiplier, customDrivable.getDirtMultiplier);
    self.setCruiseControlState = customDrivable.setCruiseControlState;
    self.setCruiseControlMaxSpeed = customDrivable.setCruiseControlMaxSpeed;
    self.toggleLowerAllImplements = customDrivable.toggleLowerAllImplements;
    self.lowerImplementByJointIndex = customDrivable.lowerImplementByJointIndex;
    self.setMirrorVisible = customDrivable.setMirrorVisible;
    self.loadAttacherJointFromXML = Utils.overwrittenFunction(self.loadAttacherJointFromXML, customDrivable.loadAttacherJointFromXML);
end

---Called on loading
-- @param table savegame savegame
-- @includeCode
function customDrivable:load(savegame)
    customDrivable.loadSettingsFromXML(self, self.xmlFile);

    self.isDrivable = true;
    self.steeringEnabled = true;
    self.nonTabbable = false

    self.axisSmoothTime = 500;
    self.axisForward = 0;
    self.lastDigitalForward = 0;
    self.axisForwardIsAnalog = false;
    self.axisSide = 0;
    self.axisSideIsAnalog = false;
    self.doHandbrake = false;
    self.reverserDirection = 1;

    self.tickDt = 30;

    -- mirrors
    self.mirrors = {};
    local useMirrors = g_gameSettings:getValue("maxNumMirrors") > 0;
    local i = 0;
    while true do
        local key = string.format("vehicle.mirrors.mirror(%d)", i);
        if not hasXMLProperty(self.xmlFile, key) then
            break;
        end
        local node = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#index"));
        if node ~= nil then
            local prio = Utils.getNoNil(getXMLInt(self.xmlFile, key.."#prio"), 2)
            setReflectionMapObjectMasks(node, 2147483776, 2155872256, true); --0x80000080, 0x80800000
            if getObjectMask(node) == 0 then
                -- Mirrors should not be visible in other mirrors
                setObjectMask(node, 16711807); -- 0x00FF007F
            end

            if useMirrors then
                table.insert(self.mirrors, {node=node, prio=prio, cosAngle=1});
            else
                setVisibility(node, false)
            end
        end
        i = i + 1;
    end

    self:setMirrorVisible(self.cameras[self.camIndex].useMirror);

    self.cruiseControl = {};
    self.cruiseControl.maxSpeed = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.cruiseControl#maxSpeed"), math.ceil(self.motor:getMaximumForwardSpeed()*3.6));
    self.cruiseControl.minSpeed = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.cruiseControl#minSpeed"), math.min(1, self.cruiseControl.maxSpeed));
    self.cruiseControl.speed = self.cruiseControl.maxSpeed;
    self.cruiseControl.isActive = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.cruiseControl#enabled"), true);
    self.cruiseControl.state = customDrivable.CRUISECONTROL_STATE_OFF;
    self.cruiseControl.topSpeedTime = 2000;
    self.cruiseControl.changeDelay = 250;
    self.cruiseControl.changeCurrentDelay = 0;
    self.cruiseControl.changeMultiplier = 1;
    self.cruiseControl.speedSent = self.cruiseControl.speed;

    self.lastOperatingTime = 0;
    self.maxOperatingTime = 99999.9 * 1000 * 60 * 60;

    self.operatingTimeHud = VehicleHudUtils.loadHud(self, self.xmlFile, "operatingTime");
    self.cruiseControlHud = VehicleHudUtils.loadHud(self, self.xmlFile, "cruiseControl");

    if self.cruiseControlHud ~= nil then
        VehicleHudUtils.setHudValue(self, self.cruiseControlHud, self.cruiseControl.speed, 9999);
    end

    if self.isClient then
        self.turnLightRepetitionCount = 0;

        local useDefaultSample = true;
        if hasXMLProperty(self.xmlFile, "vehicle.turnLightSound") then
            useDefaultSample = false;
            local customFile = getXMLString(self.xmlFile, "vehicle.turnLightSound#file");
            local customFilename = Utils.getFilename(customFile, self.baseDirectory);
            useDefaultSample = g_currentMission.defaultVehicleSounds[customFilename] ~= nil;
        end
        if not useDefaultSample then
            self.sampleTurnLight = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.turnLightSound", defaultSample, self.baseDirectory);
        end

        local useDefaultSample = true;
        if hasXMLProperty(self.xmlFile, "vehicle.waterSplashSound") then
            useDefaultSample = false;
            local customFile = getXMLString(self.xmlFile, "vehicle.waterSplashSound#file");
            local customFilename = Utils.getFilename(customFile, self.baseDirectory);
            useDefaultSample = g_currentMission.defaultVehicleSounds[customFilename] ~= nil;
        end
        if not useDefaultSample then
            self.sampleWaterSplash = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.waterSplashSound", defaultSample, self.baseDirectory);
        end

    end

    self.drivableGroundFlag = self:getNextDirtyFlag();

    if savegame ~= nil then
        local maxSpeed = getXMLInt(savegame.xmlFile, savegame.key.."#cruiseControl");
        self:setCruiseControlMaxSpeed(maxSpeed);

        local dir = getXMLInt(savegame.xmlFile, savegame.key.."#attacherJointComboDir");
        if dir ~= nil then
            self.attacherJointCombos.direction = dir;
            if dir == 1 then
                self.attacherJointCombos.currentTime = self.attacherJointCombos.duration;
            end
        end
    end

    self.showChangeToolSelectionHelp = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.showChangeToolSelectionHelp"), true)
    self.showToolSelectionHud = true;
end

---Called on loading
-- @param table savegame savegame
-- @includeCode
function customDrivable:loadSettingsFromXML(xmlFile)
    self.speedRotScale = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.speedRotScale#scale"), 80);
    self.speedRotScaleOffset = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.speedRotScale#offset"), 0.7);
    self.maxRotatedTimeSpeed = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.maxRotatedTimeSpeed#value"), 1)*0.001;
end

---Called on deleting
-- @includeCode
function customDrivable:delete()
    if self.sampleTurnLight ~= nil then
        SoundUtil.deleteSample(self.sampleTurnLight);
    end
    if self.sampleWaterSplash ~= nil then
        SoundUtil.deleteSample(self.sampleWaterSplash);
    end
end

---Called on client side on join
-- @param integer streamId streamId
-- @param integer connection connection
-- @includeCode
function customDrivable:readStream(streamId, connection)
    if self.cruiseControl.isActive then
        self:setCruiseControlState(streamReadUIntN(streamId, 2), true);
        local speed = streamReadUInt8(streamId);
        self:setCruiseControlMaxSpeed(speed);
        self.cruiseControl.speedSent = speed;
    end

    if self.attacherJointCombos ~= nil then
        self.attacherJointCombos.direction = -1;
        local dir = streamReadBool(streamId);
        if dir then
            self.attacherJointCombos.direction = 1;
            self.attacherJointCombos.currentTime = self.attacherJointCombos.duration;
        end
    end
end

---Called on server side on join
-- @param integer streamId streamId
-- @param integer connection connection
-- @includeCode
function customDrivable:writeStream(streamId, connection)
    if self.cruiseControl.isActive then
        streamWriteUIntN(streamId, self.cruiseControl.state, 2);
        streamWriteUInt8(streamId, self.cruiseControl.speed);
    end

    if self.attacherJointCombos ~= nil then
        streamWriteBool(streamId, self.attacherJointCombos.direction == 1);
    end
end

---Called on on update
-- @param integer streamId stream ID
-- @param integer timestamp timestamp
-- @param table connection connection
-- @includeCode
function customDrivable:readUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        local brakeLightsVisibility = streamReadBool(streamId);
        self:setBrakeLightsVisibility(brakeLightsVisibility);
        local reverseLightsVisibility = streamReadBool(streamId);
        self:setReverseLightsVisibility(reverseLightsVisibility);
    else
        local hasUpdate = streamReadBool(streamId);
        if hasUpdate then
            local axisForwardIsAnalog = streamReadBool(streamId);
            local axisSideIsAnalog = streamReadBool(streamId);
            local axisForward = streamReadFloat32(streamId);
            local axisSide = streamReadFloat32(streamId);
            local doHandbrake = streamReadBool(streamId);
            local dt = streamReadFloat32(streamId);
            if self.steeringEnabled and self.isControlled then
                customDrivable.updateVehiclePhysics(self, axisForward, axisForwardIsAnalog, axisSide, axisSideIsAnalog, doHandbrake, dt);
            end
        end
    end
end

---Called on on update
-- @param integer streamId stream ID
-- @param table connection connection
-- @param integer dirtyMask dirty mask
-- @includeCode
function customDrivable:writeUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        streamWriteBool(streamId, self.brakeLightsVisibility);
        streamWriteBool(streamId, self.reverseLightsVisibility);
    else
        if streamWriteBool(streamId, bitAND(dirtyMask, self.drivableGroundFlag) ~= 0) then
            streamWriteBool(streamId, self.axisForwardIsAnalog);
            streamWriteBool(streamId, self.axisSideIsAnalog);
            streamWriteFloat32(streamId, self.axisForward);
            streamWriteFloat32(streamId, self.axisSide);
            streamWriteBool(streamId, self.doHandbrake);
            streamWriteFloat32(streamId, self.tickDt);
        end
    end
end

---Called on mouse event
-- @param float posX x position
-- @param float posY y position
-- @param boolean isDown is mouse button down
-- @param boolean isUp is mouse button up
-- @param integer button mouse button
-- @includeCode
function customDrivable:mouseEvent(posX, posY, isDown, isUp, button)
end

---Called on key event
-- @param integer unicode unicode
-- @param integer sym sym
-- @param integer modifier modifier
-- @param boolean isDown is button down
-- @includeCode
function customDrivable:keyEvent(unicode, sym, modifier, isDown)
end

---Called on update
-- @param float dt time since last call in ms
-- @includeCode
function customDrivable:update(dt)
    if self.isEntered and self.isClient then

        local turnLightRepetitionCount = math.floor((getShaderTimeSec()*7 + math.acos(-0.2)) / (math.pi*2));
        if self.turnLightRepetitionCount ~= nil and turnLightRepetitionCount ~= self.turnLightRepetitionCount then
            if self.turnLightState > Lights.TURNLIGHT_OFF then
                if self.sampleTurnLight then
                    SoundUtil.playSample(self.sampleTurnLight, 1, 0, nil);
                else
                    SoundUtil.playSample(g_currentMission.sampleTurnLight, 1, 0, nil);
                end
            end
        end
        self.turnLightRepetitionCount = turnLightRepetitionCount;

        --
        if self.cruiseControl.isActive then
            -- Cruise Control
            if self:getIsActiveForInput(false, false) then
                if InputBinding.hasEvent(InputBinding.TOGGLE_CRUISE_CONTROL) then
                    if self.cruiseControl.state == customDrivable.CRUISECONTROL_STATE_OFF then
                        self:setCruiseControlState(customDrivable.CRUISECONTROL_STATE_ACTIVE);
                    else
                        self:setCruiseControlState(customDrivable.CRUISECONTROL_STATE_OFF);
                    end
                    self.cruiseControl.topSpeedTime = 500;
                elseif InputBinding.isPressed(InputBinding.TOGGLE_CRUISE_CONTROL) then
                    self.cruiseControl.topSpeedTime = self.cruiseControl.topSpeedTime - dt;
                    if self.cruiseControl.topSpeedTime < 0 then
                        self:setCruiseControlState(customDrivable.CRUISECONTROL_STATE_FULL);
                    end
                else
                    self.cruiseControl.topSpeedTime = 500;
                end
            end

            if not (g_gui:getIsGuiVisible() or g_currentMission.isPlayerFrozen or g_currentMission.inGameMessage:getIsVisible()) then --and self.selectedImplement == nil then
                -- Cruise Control speed
                local inputW = InputBinding.getDigitalInputAxis(InputBinding.AXIS_CRUISE_CONTROL);
                if InputBinding.isAxisZero(inputW) then
                    inputW = InputBinding.getAnalogInputAxis(InputBinding.AXIS_CRUISE_CONTROL)
                end

                if inputW ~= 0 then
                    self.cruiseControl.changeCurrentDelay = self.cruiseControl.changeCurrentDelay - dt*self.cruiseControl.changeMultiplier;
                    self.cruiseControl.changeMultiplier = math.min(self.cruiseControl.changeMultiplier + dt*0.003, 10);
                    if self.cruiseControl.changeCurrentDelay < 0 then
                        local dir = 0;
                        if inputW > g_analogStickVTolerance then
                            dir = 1;
                        elseif inputW < -g_analogStickVTolerance then
                            dir = -1;
                        end

                        if dir ~= 0 then
                            self.cruiseControl.wasSpeedChanged = true;
                            self:setCruiseControlMaxSpeed(self.cruiseControl.speed + dir);
                            self.cruiseControl.changeCurrentDelay = self.cruiseControl.changeDelay;
                        end
                    end
                else
                    self.cruiseControl.wasSpeedChanged = false;
                    self.cruiseControl.changeMultiplier = 1;
                    self.cruiseControl.changeCurrentDelay = 0;
                end

                if not self.cruiseControl.wasSpeedChanged and self.cruiseControl.speed ~= self.cruiseControl.speedSent then
                    if g_server ~= nil then
                        g_server:broadcastEvent(SetCruiseControlSpeedEvent:new(self, self.cruiseControl.speed), nil, nil, self);
                    else
                        g_client:getServerConnection():sendEvent(SetCruiseControlSpeedEvent:new(self, self.cruiseControl.speed));
                    end
                    self.cruiseControl.speedSent = self.cruiseControl.speed;
                end

            end
        end

        if not self.isHired then
            -- do all the input handling
            if self:getIsActiveForInput(false, false) then

                if InputBinding.hasEvent(InputBinding.TOGGLE_TIPSTATE) then
                    self:handleToggleTipStateEvent();
                end

                if InputBinding.hasEvent(InputBinding.ATTACH) then
                    self:handleAttachEvent();
                end

                if InputBinding.hasEvent(InputBinding.LOWER_IMPLEMENT) then
                    self:handleLowerImplementEvent();
                end

                if InputBinding.hasEvent(InputBinding.LOWER_ALL_IMPLEMENTS) then
                    self:toggleLowerAllImplements();
                end

                if InputBinding.hasEvent(InputBinding.SWITCH_IMPLEMENT) then
                    self:selectNextSelectableImplement();
                end

                if self.hasTurnLights then
                    if InputBinding.hasEvent(InputBinding.TOGGLE_TURNLIGHT_HAZARD) then
                        local state = Lights.TURNLIGHT_OFF;
                        if self.turnLightState ~= Lights.TURNLIGHT_HAZARD then
                            state = Lights.TURNLIGHT_HAZARD;
                        end
                        self:setTurnLightState(state);
                    end
                    if InputBinding.hasEvent(InputBinding.TOGGLE_TURNLIGHT_LEFT) then
                        local state = Lights.TURNLIGHT_OFF;
                        if self.turnLightState ~= Lights.TURNLIGHT_LEFT then
                            state = Lights.TURNLIGHT_LEFT;
                        end
                        self:setTurnLightState(state);
                    end
                    if InputBinding.hasEvent(InputBinding.TOGGLE_TURNLIGHT_RIGHT) then
                        local state = Lights.TURNLIGHT_OFF;
                        if self.turnLightState ~= Lights.TURNLIGHT_RIGHT then
                            state = Lights.TURNLIGHT_RIGHT;
                        end
                        self:setTurnLightState(state);
                    end
                end

                if InputBinding.hasEvent(InputBinding.TOGGLE_BEACON_LIGHTS) then
                    self:setBeaconLightsVisibility(not self.beaconLightsActive);
                end

            end

            if self:getIsActiveForInput(false, false) then

                self.doHandbrake = false;
                local axisAccelerate = InputBinding.getDigitalInputAxis(InputBinding.AXIS_ACCELERATE_VEHICLE);
                local axisBrake = InputBinding.getDigitalInputAxis(InputBinding.AXIS_BRAKE_VEHICLE)
                local axisForward = Utils.clamp((axisAccelerate-axisBrake)*0.5, -1, 1);

                if InputBinding.isAxisZero(axisForward) then
                    axisAccelerate = InputBinding.getAnalogInputAxis(InputBinding.AXIS_ACCELERATE_VEHICLE);
                    axisBrake = InputBinding.getAnalogInputAxis(InputBinding.AXIS_BRAKE_VEHICLE)
                    self.axisForward = Utils.clamp((axisAccelerate-axisBrake)*0.5, -1, 1);
                    if not InputBinding.isAxisZero(self.axisForward) then
                        self.axisForwardIsAnalog = true;
                    end
                    self.lastDigitalForward = 0;
                    if math.abs(self.axisForward) > 0.1 and g_gui.currentGuiName ~= "ChatDialog" then
                        self:setCruiseControlState(customDrivable.CRUISECONTROL_STATE_OFF);
                    end
                else
                    self.axisForward = axisForward;
                    self.axisForwardIsAnalog = false;
                    self.lastDigitalForward = self.axisForward;
                    if g_gui.currentGuiName ~= "ChatDialog" then
                        self:setCruiseControlState(customDrivable.CRUISECONTROL_STATE_OFF);
                    end
                end
            else
                self.doHandbrake = true;
            end
        else
            self.axisForward = 0;
            self.axisForwardIsAnalog = false;
        end

        -- do mirror checks
        if self.activeCamera ~= nil and self.activeCamera.useMirror then
            self:setMirrorVisible(true)
        end

        if not self.isHired then
            self.axisSide = InputBinding.getDigitalInputAxis(InputBinding.AXIS_MOVE_SIDE_VEHICLE);
            if InputBinding.isAxisZero(self.axisSide) then
                self.axisSide = InputBinding.getAnalogInputAxis(InputBinding.AXIS_MOVE_SIDE_VEHICLE)
                if not InputBinding.isAxisZero(self.axisSide) then
                    self.axisSideIsAnalog = true;
                end
            else
                self.axisSideIsAnalog = false;
            end

            if not self:getIsActiveForInput(false, true) then
                if not self.axisSideIsAnalog or g_gui.currentGuiName == "PlacementScreen" then
                    self.axisSide = 0;
                end
                if not self.axisForwardIsAnalog or g_gui.currentGuiName == "PlacementScreen" then
                    self.axisForward = 0;
                end

                if self.steeringEnabled then
                    if g_gui:getIsGuiVisible() and g_gui.currentGuiName ~= "ChatDialog" then
                        self:setCruiseControlState(customDrivable.CRUISECONTROL_STATE_OFF);
                    end
                end
            end

            if g_isServerStreamingVersion then self.axisSide = self.axisSide * 0.5; end   -- This is the factor to slow down the steering
        end

        if self.isServer then
            if self.steeringEnabled then
                customDrivable.updateVehiclePhysics(self, self.axisForward, self.axisForwardIsAnalog, self.axisSide, self.axisSideIsAnalog, self.doHandbrake, dt);
            end
        else
            self:raiseDirtyFlags(self.drivableGroundFlag);
        end
    end

    if self:getIsActive() then
        -- lock max speed to working tool
        local speed,_ = self:getSpeedLimit(true);
        if self.cruiseControl.state == customDrivable.CRUISECONTROL_STATE_ACTIVE then
            speed = math.min(speed, self.cruiseControl.speed);
        end
        self.motor:setSpeedLimit(speed);
    end
end

---Called on update tick
-- @param float dt time since last call in ms
-- @includeCode
function customDrivable:updateTick(dt)
    self.tickDt = dt;
    if self:getIsActive() then
        local xt,yt,zt = getTranslation(self.components[1].node);
        local deltaWater = yt-g_currentMission.waterY+2.5;
        if deltaWater < 0 then
            self.isBroken = true;
            g_currentMission:onSunkVehicle(self);

            if self.isEntered then
                if self:getIsActiveForSound() then
                    local volume = math.min(1, self:getLastSpeed()/30);
                    if self.sampleWaterSplash ~= nil then
                        SoundUtil.playSample(self.sampleWaterSplash, 1, 0, volume);
                    else
                        SoundUtil.playSample(g_currentMission.sampleWaterSplash, 1, 0, volume);
                    end
                end
                g_currentMission:onLeaveVehicle();
            end
        end
        self.showWaterWarning = deltaWater < 2;

        if self.attacherJointCombos ~= nil and self.attacherJointCombos.isRunning then
            for k, joint in pairs(self.attacherJointCombos.joints) do
                if self.attacherJointCombos.direction == 1 and self.attacherJointCombos.currentTime >= joint.time then
                    self:lowerImplementByJointIndex(joint.jointIndex, true);
                elseif self.attacherJointCombos.direction == -1 and self.attacherJointCombos.currentTime <= self.attacherJointCombos.duration-joint.time then
                    self:lowerImplementByJointIndex(joint.jointIndex, false);
                end
            end

            if (self.attacherJointCombos.direction == -1 and self.attacherJointCombos.currentTime == 0) or
               (self.attacherJointCombos.direction == 1  and self.attacherJointCombos.currentTime == self.attacherJointCombos.duration) then
                self.attacherJointCombos.isRunning = false;
            end

            self.attacherJointCombos.currentTime = Utils.clamp(self.attacherJointCombos.currentTime + dt*self.attacherJointCombos.direction, 0, self.attacherJointCombos.duration);
        end
    end
end

---Called on draw
-- @includeCode
function customDrivable:draw()

    if not self.isEntered then
        return;
    end

    -- mirror debug rendering
    --[[for k, mirror in pairs(self.mirrors) do
        setTextAlignment(RenderText.ALIGN_LEFT);
        renderText(0.5, 0.9-(k-1)*0.015, 0.012, string.format("%s\t: %s %.4f", getName(mirror.node), tostring(getVisibility(mirror.node)), math.deg(math.acos(mirror.cosAngle))))

        local x,y,z = getWorldTranslation(mirror.node)
        Utils.renderTextAtWorldPosition(x,y,z, string.format("%s: %.3f", getName(mirror.node), math.deg(math.acos(mirror.cosAngle))), 0.015, 0)
    end]]

    g_currentMission:drawVehicleHud(self);

    if self.showWaterWarning then
        g_currentMission:showBlinkingWarning(g_i18n:getText("warning_dontDriveIntoWater"), 2000);
    end

    if not self.isHired then
        if self == g_currentMission.controlledVehicle then
            if g_currentMission.trailerInTipRange ~= nil and (g_currentMission.trailerInTipRange.getFillLevel ~= nil and g_currentMission.trailerInTipRange:getFillLevel() > 0) then
                g_currentMission:enableHudIcon("tip", 4);
                g_currentMission:addHelpButtonText(g_currentMission.trailerInTipRange.tipText, InputBinding.TOGGLE_TIPSTATE, nil, GS_PRIO_VERY_HIGH);
            end

            -- enable attachment icon if needed
            if (g_currentMission.attachableInMountRange ~= nil and g_currentMission.attachableInMountRangeVehicle.attacherJoints[g_currentMission.attachableInMountRangeIndex].jointIndex == 0) then
                g_currentMission:enableHudIcon("attach", 3);
                g_currentMission:addHelpButtonText(g_i18n:getText("action_attach"), InputBinding.ATTACH, nil, GS_PRIO_VERY_HIGH);
            else
                -- if something is detachable is attached, show button help
                if self:detachingIsPossible() then
                    g_currentMission:addHelpButtonText(g_i18n:getText("action_detach"), InputBinding.ATTACH, nil, GS_PRIO_VERY_HIGH);
                end
            end

            if self.selectedImplement ~= nil and self.selectedImplement.object ~= nil and self.selectedImplement.object.attacherVehicle ~= nil and not self.isHired then
                local object = self.selectedImplement.object;
                local jointDesc = object.attacherVehicle.attacherJoints[self.selectedImplement.jointDescIndex];

                if object.attacherJoint.allowsLowering and jointDesc.allowsLowering and object.foldMiddleAnimTime == nil then
                    if jointDesc.moveDown then
                        g_currentMission:addHelpButtonText(string.format(g_i18n:getText("action_liftOBJECT"), object.typeDesc), InputBinding.LOWER_IMPLEMENT, nil, GS_PRIO_HIGH);
                    else
                        g_currentMission:addHelpButtonText(string.format(g_i18n:getText("action_lowerOBJECT"), object.typeDesc), InputBinding.LOWER_IMPLEMENT, nil, GS_PRIO_HIGH);
                    end
                end
            end
        end

        if #self.beaconLights > 0 then
            g_currentMission:addHelpButtonText(g_i18n:getText("input_TOGGLE_BEACON_LIGHTS"), InputBinding.TOGGLE_BEACON_LIGHTS, nil, GS_PRIO_VERY_LOW);
        end

        if self.cruiseControl.state == customDrivable.CRUISECONTROL_STATE_OFF then
            g_currentMission:addHelpButtonText(g_i18n:getText("action_activateCruiseControl"), InputBinding.TOGGLE_CRUISE_CONTROL, nil, GS_PRIO_LOW);
        else
            g_currentMission:addHelpButtonText(g_i18n:getText("action_deactivateCruiseControl"), InputBinding.TOGGLE_CRUISE_CONTROL, nil, GS_PRIO_LOW);
        end

        g_currentMission:addHelpButtonText(g_i18n:getText("action_changeCruiseControlLevel"), InputBinding.AXIS_CRUISE_CONTROL, nil, GS_PRIO_LOW);

        local numDirectImplements = table.getn(self.attachedImplements);
        local hasMoreThan1Selectable = (numDirectImplements > 1) or (self:getIsSelectable() and numDirectImplements > 0);
        if not hasMoreThan1Selectable then
            for _, implement in pairs(self.attachedImplements) do
                if implement.object ~= nil and table.getn(implement.object.attachedImplements) > 0 then
                    hasMoreThan1Selectable = true;
                    break;
                end
            end
        end

        if hasMoreThan1Selectable and self.showChangeToolSelectionHelp then
            g_currentMission:addHelpButtonText(g_i18n:getText("action_changeToolSelection"), InputBinding.SWITCH_IMPLEMENT, nil, GS_PRIO_NORMAL);
        end
    end

    if GS_IS_CONSOLE_VERSION then
        g_currentMission:addHelpButtonText(g_i18n:getText("input_RADIO_TOGGLE"), InputBinding.RADIO_TOGGLE, nil, GS_PRIO_VERY_LOW);
        if g_gameSettings:getValue("radioIsActive") then
            g_currentMission:addHelpButtonText(g_i18n:getText("input_RADIO_NEXT_CHANNEL"), InputBinding.RADIO_NEXT_CHANNEL, nil, GS_PRIO_VERY_LOW);
            g_currentMission:addHelpButtonText(g_i18n:getText("input_RADIO_PREVIOUS_CHANNEL"), InputBinding.RADIO_PREVIOUS_CHANNEL, nil, GS_PRIO_VERY_LOW);
        end
    end
end

---Called on enter vehicle
-- @param boolean isControlling is player controlling the vehicle
-- @includeCode
function customDrivable:onEnter(isControlling)
    self:onActivateAttachments();
end

---Called on leaving the vehicle
-- @includeCode
function customDrivable:onLeave()
    --if self:getDeactivateOnLeave() then
        if self.isServer then
            if next(self.differentials) ~= nil and self.motorizedNode ~= nil then
                setVehicleProps(self.motorizedNode, 0.0, 0.0, 0.0, self.motor:getMaxClutchTorque(), self.motor:getRotInertia(), self.motor:getDampingRate());
            end
            for _,wheel in pairs(self.wheels) do
                setWheelShapeProps(wheel.node, wheel.wheelShape, 0, self.motor:getBrakeForce()*wheel.brakeFactor, 0, wheel.rotationDamping);
            end
        end
    -- end

    self:setCruiseControlState(customDrivable.CRUISECONTROL_STATE_OFF);

    self:setMirrorVisible(false);
end

---Changes cruise control state
-- @param integer state new state
-- @param boolean noEventSend no event send
-- @includeCode
function customDrivable:setCruiseControlState(state, noEventSend)
    if self.cruiseControl.state ~= state then
        self.cruiseControl.state = state;
        if not self.isServer and (noEventSend == nil or not noEventSend) then
            g_client:getServerConnection():sendEvent(SetCruiseControlStateEvent:new(self, state));
        end
    end
end

---Sets cruise control max speed
-- @param float speed new max speed
-- @includeCode
function customDrivable:setCruiseControlMaxSpeed(speed)
    if speed ~= nil then
        speed = Utils.clamp(speed, self.cruiseControl.minSpeed, self.cruiseControl.maxSpeed);
        if self.cruiseControl.speed ~= speed then
            self.cruiseControl.speed = speed;
            if self.cruiseControl.state == customDrivable.CRUISECONTROL_STATE_FULL then
                self.cruiseControl.state = customDrivable.CRUISECONTROL_STATE_ACTIVE;
            end
        end

        if self.cruiseControl.state ~= customDrivable.CRUISECONTROL_STATE_OFF then
            if self.cruiseControl.state == customDrivable.CRUISECONTROL_STATE_ACTIVE then
                self.motor:setSpeedLimit(self.cruiseControl.speed);
            else
                self.motor:setSpeedLimit(math.huge);
            end
        end

        if self.cruiseControlHud ~= nil then
            VehicleHudUtils.setHudValue(self, self.cruiseControlHud, self.cruiseControl.speed, 9999);
        end
    end

end

---Returns rpm limit
-- @includeCode
function customDrivable:getRpmLimit(superFunc)
    if superFunc ~= nil then
        return superFunc(self);
    end
    return math.huge;
end

---Returns current dirt multiplier
-- @return float dirtMultiplier current dirt multiplier
-- @includeCode
function customDrivable:getDirtMultiplier(superFunc)
    local multiplier = 0;
    if superFunc ~= nil then
        multiplier = multiplier + superFunc(self);
    end

    return multiplier;
end

---Changes visibility of mirrors
-- @param boolean visible visible
-- @includeCode
function customDrivable:setMirrorVisible(visible)
    if next(self.mirrors) == nil then
        return
    end
    if visible then
        local numVisibleMirrors = 0
        for _, mirror in pairs(self.mirrors) do
            -- only use mirrors which are visible on screen
            if getIsInCameraFrustum(mirror.node, self.activeCamera.cameraNode, g_presentedScreenAspectRatio) then
                -- calculate angle between mirror and camera
                local dirX, dirY, dirZ = worldToLocal(self.activeCamera.cameraNode, getWorldTranslation(mirror.node))

                dirY = dirY * g_screenAspectRatio
                local length = Utils.vector3Length(dirX, dirY, dirZ)
                mirror.cosAngle = -dirZ / length
            else
                mirror.cosAngle = math.huge
            end
        end

        -- sort mirrors based on prio and angle
        table.sort(self.mirrors,
                   function(mirror1, mirror2)
                        if mirror1.prio == mirror2.prio then
                            -- the bigger cosAngle, the smaller the angle to the z-axis
                            return mirror1.cosAngle > mirror2.cosAngle
                        else
                            return mirror1.prio < mirror2.prio
                        end
                   end)

        local maxNumMirrors = g_gameSettings:getValue("maxNumMirrors")
        -- show first mirrors within the limit
        for _, mirror in ipairs(self.mirrors) do
            if mirror.cosAngle ~= math.huge and numVisibleMirrors < maxNumMirrors then
                setVisibility(mirror.node, true)
                numVisibleMirrors = numVisibleMirrors + 1
            else
                setVisibility(mirror.node, false)
            end
        end
    else
        for _, mirror in pairs(self.mirrors) do
            setVisibility(mirror.node, false);
        end
    end
end

---Called on camera change
-- @param table camera new camera
-- @param integer camIndex index of new camera
-- @includeCode
function customDrivable:onCameraChanged(camera, camIndex)
    self:setMirrorVisible(camera.useMirror);
end

---Called on motor stop
-- @param boolean noEventSend no event send
-- @includeCode
function customDrivable:stopMotor(noEventSend)
    self:setCruiseControlState(customDrivable.CRUISECONTROL_STATE_OFF, true);
end

---Returns attributes and nodes to save
-- @param table nodeIdent node ident
-- @return string attributes attributes
-- @return string nodes nodes
-- @includeCode
function customDrivable:getSaveAttributesAndNodes(nodeIdent)
    local attributes = 'cruiseControl="'..tostring(self.cruiseControl.speed)..'" ';

    if self.attacherJointCombos ~= nil then
        attributes = attributes .. ' attacherJointComboDir="'..tostring(self.attacherJointCombos.direction)..'"';
    end

    local nodes = "";

    return attributes, nodes;
end

---Update vehicle physics
-- @param float axisForward axis forward value
-- @param boolean axisForwardIsAnalog forward axis is analog
-- @param float axisSide side axis
-- @param boolean axisSideIsAnalog side axis is analog
-- @param boolean doHandbrake do handbrake
-- @param float dt time since lsat call in ms
-- @includeCode
function customDrivable.updateVehiclePhysics(self, axisForward, axisForwardIsAnalog, axisSide, axisSideIsAnalog, doHandbrake, dt)

    axisForward = axisForward;
    axisSide = self.reverserDirection * axisSide;

    local acceleration = 0;
    if self.isMotorStarted and self.motorStartTime <= g_currentMission.time then
        acceleration = -axisForward;
        if math.abs(acceleration) > 0.8 then
            self:setCruiseControlState(customDrivable.CRUISECONTROL_STATE_OFF, true);
        end
        if self.cruiseControl.state ~= customDrivable.CRUISECONTROL_STATE_OFF then
            acceleration = 1.0;
        end
    end
    if self.fuelFillLevel == 0 then
        acceleration = 0;
    end

    local inputAxisX = axisSide;
    if axisSideIsAnalog then
        local targetRotatedTime = 0;
        if inputAxisX < 0 then
            -- 0 to maxRotTime
            targetRotatedTime = math.min(-self.maxRotTime*inputAxisX, self.maxRotTime);
        else
            -- 0 to minRotTime
            targetRotatedTime = math.max(self.minRotTime*inputAxisX, self.minRotTime);
        end
        local maxTime = self.maxRotatedTimeSpeed*dt;
        if math.abs(targetRotatedTime-self.rotatedTime) > maxTime then
            if targetRotatedTime > self.rotatedTime then
                self.rotatedTime = self.rotatedTime + maxTime;
            else
                self.rotatedTime = self.rotatedTime - maxTime;
            end
        else
            self.rotatedTime = targetRotatedTime;
        end
    else
        local rotScale = math.min(1.0/(self.lastSpeed*self.speedRotScale+self.speedRotScaleOffset), 1);
        if inputAxisX < 0 then
            self.rotatedTime = math.min(self.rotatedTime - dt*0.001*inputAxisX*rotScale, self.maxRotTime);
        elseif inputAxisX > 0 then
            self.rotatedTime = math.max(self.rotatedTime - dt*0.001*inputAxisX*rotScale, self.minRotTime);
        else
            if self.autoRotateBackSpeed ~= 0 then
                if self.rotatedTime > 0 then
                    self.rotatedTime = math.max(self.rotatedTime - dt*0.001*self.autoRotateBackSpeed*rotScale, 0);
                else
                    self.rotatedTime = math.min(self.rotatedTime + dt*0.001*self.autoRotateBackSpeed*rotScale, 0);
                end
            end
        end
    end

    if self.firstTimeRun then
        if self.motorType == "locomotive" then
            self.motor:updateInput(dt, acceleration);
        else
            self.motor:setRpmLimit( self:getRpmLimit() );
            WheelsUtil.updateWheelsPhysics(self, dt, self.lastSpeedReal, acceleration, doHandbrake, self.requiredDriveMode)
        end
    end
end

---Lowers all implements
-- @param boolean noEventSend no event send
-- @includeCode
function customDrivable:toggleLowerAllImplements(noEventSend)
    if self.attacherJointCombos ~= nil then
        DrivableToggleLowerAllEvent.sendEvent(self, noEventSend);
        self.attacherJointCombos.direction = self.attacherJointCombos.direction * -1;
        self.attacherJointCombos.isRunning = true;
    end
end

---Called on reverse direction change
-- @param integer direction direction [-1, 0, 1]
-- @includeCode
function customDrivable:onReverseDirectionChanged(direction)
    if self.attacherJointCombos ~= nil then
        for k, joint in pairs(self.attacherJointCombos.joints) do
            joint.time = math.abs(joint.time - self.attacherJointCombos.duration)
        end
    end
end

---Lower implement by joint desc index
-- @param integer jointDescIndex index of attacher joint
-- @param boolean doLowering do lowering
-- @includeCode
function customDrivable:lowerImplementByJointIndex(jointDescIndex, doLowering)
    local implementIndex = self:getImplementIndexByJointDescIndex(jointDescIndex);
    if implementIndex ~= nil then
        local implement = self.attachedImplements[implementIndex];
        if implement ~= nil then
            implement.object:onLowerAll(doLowering, jointDescIndex);
        end
    end
end

---Returns if vehicle is operating
-- @return boolean isOperating is operating
-- @includeCode
function customDrivable:getIsOperating(superFunc)
    local isOperating = true
    if superFunc ~= nil then
        isOperating = superFunc(self)
    end

    return isOperating or self.isHired
end

---Set operating time
-- @param float operatingTime operating time in ms
-- @param boolean isLoading function called on loading
-- @includeCode
function customDrivable:setOperatingTime(superFunc, operatingTime, isLoading)
    if superFunc ~= nil then
        superFunc(self, operatingTime, isLoading)
    end;

    if self.operatingTimeHud ~= nil and math.abs(self.lastOperatingTime-self.operatingTime) > 1000*60 then
        local operatingTime = self.operatingTime
        if operatingTime > self.maxOperatingTime then
            operatingTime = 0;
        end

        local minutes = operatingTime / (1000 * 60);
        local hours = math.floor(minutes / 60);
        minutes = math.floor(minutes - hours * 60);
        local minutesString = string.format("%02d", minutes);
        VehicleHudUtils.setHudValue(self, self.operatingTimeHud, tonumber(hours.."."..minutesString), self.maxOperatingTime);
        self.lastOperatingTime = self.operatingTime;
    end
end

---Load attacher joint from xml
-- @param table attacherJoint attacherJoint
-- @param integer xmlFile id of xml object
-- @param string baseName baseName
-- @param integer index index of attacher joint
-- @return boolean success success
-- @includeCode
function customDrivable:loadAttacherJointFromXML(superFunc, attacherJoint, xmlFile, baseName, index)
    if superFunc ~= nil then
        if not superFunc(self, attacherJoint, xmlFile, baseName, index) then
            return false;
        end;
    end;

    if self.attacherJointCombos == nil then
        self.attacherJointCombos = {}
        self.attacherJointCombos.duration = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.attacherJoints#comboDuration"), 2) * 1000;
        self.attacherJointCombos.currentTime = 0;
        self.attacherJointCombos.direction = -1;
        self.attacherJointCombos.isRunning = false;
        self.attacherJointCombos.joints = {};
    end

    local t = getXMLFloat(xmlFile, baseName .. "#comboTime")
    if t ~= nil then
        table.insert(self.attacherJointCombos.joints, {jointIndex=index+1, time=Utils.clamp(t, 0, 1)*self.attacherJointCombos.duration});
    end

    return true;
end
