--
-- Steerable
--
-- @author  Stefan Geiger
-- @date  30/11/08
--
-- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.

---Specialization class for steerables
-- @category Specializations
customSteerable = {};

---Checks if all prerequisite specializations are loaded
-- @param table specializations specializations
-- @return boolean hasPrerequisite true if all prerequisite specializations are loaded
-- @includeCode

function customSteerable.prerequisitesPresent(specializations)
  return true
  -- return SpecializationUtil.hasSpecialization(customlights, specializations);
end

---Called before loading
-- @param table savegame savegame
-- @includeCode
function customSteerable:preLoad(savegame)
    Vehicle.registerInteractionFlag("steerable")

    self.enterVehicle = customSteerable.enterVehicle
    self.leaveVehicle = customSteerable.leaveVehicle
    self.drawUIInfo = Utils.overwrittenFunction(self.drawUIInfo, customSteerable.drawUIInfo);
    self.getIsActiveForInput = Utils.overwrittenFunction(self.getIsActiveForInput, customSteerable.getIsActiveForInput);
    self.getDistanceToObject = Utils.overwrittenFunction(self.getDistanceToObject, customSteerable.getDistanceToObject);
    self.getInteractionHelp = Utils.overwrittenFunction(self.getInteractionHelp, customSteerable.getInteractionHelp);
    self.interact = Utils.overwrittenFunction(self.interact, customSteerable.interact);
    self.setActiveCameraIndex = customSteerable.setActiveCameraIndex;
    self.addToolCameras = customSteerable.addToolCameras;
    self.removeToolCameras = customSteerable.removeToolCameras;

    self.onEnter = SpecializationUtil.callSpecializationsFunction("onEnter");
    self.onLeave = SpecializationUtil.callSpecializationsFunction("onLeave");
    self.onCameraChanged = SpecializationUtil.callSpecializationsFunction("onCameraChanged");
end

---Called on loading
-- @param table savegame savegame
-- @includeCode
function customSteerable:load(savegame)

    self.isSteerable = true;
    self.isAttachable = SpecializationUtil.hasSpecialization(Attachable, self.specializations);
    self.isControlled = false;

    self.steering = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.steering#index"));

    customSteerable.loadSettingsFromXML(self, self.xmlFile);

    self.numCameras = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.cameras#count"), 0);
    self.cameras = {};
    for i=1, self.numCameras do
        local cameraKey = string.format("vehicle.cameras.camera%d", i);
        local camera = VehicleCamera:new(self);
        if camera:loadFromXML(self.xmlFile, cameraKey) then
            table.insert(self.cameras, camera);
        end
    end
    self.numCameras = table.getn(self.cameras);
    if self.numCameras == 0 then
        print("Error: No cameras in xml file: ".. self.configFileName);
    end

    self.camIndex = 1;

    self.isEntered = false;

    self.controllerName = "Unknown";

    self.disableCharacterOnLeave = true;

    self.deactivateLightsOnLeave = true;

    self.steerableNeedsAttacherVehicle = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.steerableNeedsAttacherVehicle"), true);

    self.forceSelectionOnEnter = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.forceSelectionOnEnter"), false);

    self.currentLightState = 0;
    if self.lightStates == nil or #self.lightStates == 0 then
        table.insert(self.lightStates, {0});
        table.insert(self.lightStates, {0, 1});
        table.insert(self.lightStates, {0, 1, 2});
    end

    local useDefaultSample = true;
    if hasXMLProperty(self.xmlFile, "vehicle.toggleLightsSound") then
        useDefaultSample = false;
        local customFile = getXMLString(self.xmlFile, "vehicle.toggleLightsSound#file");
        local customFilename = Utils.getFilename(customFile, self.baseDirectory);
        useDefaultSample = g_currentMission.defaultVehicleSounds[customFilename] ~= nil;
    end
    if not useDefaultSample then
        self.sampleToggleLights = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.toggleLightsSound", nil, self.baseDirectory);
    end

    self.steerableGroundFlag = self:getNextDirtyFlag();

    g_currentMission:addInteractiveVehicle(self);
end

---Called after loading
-- @param table savegame savegame
-- @includeCode
function customSteerable:postLoad(savegame)
    if self.steering ~= nil then
        self.steeringSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.steering#rotationSpeed"), 0);
        local indoorRotation = getXMLFloat(self.xmlFile, "vehicle.steering#indoorRotation");
        if indoorRotation ~= nil then
            self.steeringIndoorRotation = math.rad(indoorRotation);
        end
        local outdoorRotation = getXMLFloat(self.xmlFile, "vehicle.steering#outdoorRotation");
        if outdoorRotation ~= nil then
            self.steeringOutdoorRotation = math.rad(outdoorRotation);
        end
        _, self.steeringLastRotation, _ = getRotation(self.steering);
    end
end

---Called on loading
-- @param table savegame savegame
-- @includeCode
function customSteerable:loadSettingsFromXML(xmlFile)
    self.enterReferenceNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.enterReferenceNode#index"));
    self.exitPoint = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.exitPoint#index"));

    self.interactionRadius = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.enterReferenceNode#interactionRadius"), 6.0)

    self.vehicleCharacter = VehicleCharacter:new(self);
    if not self.vehicleCharacter:load(xmlFile, "vehicle.characterNode") then
        self.vehicleCharacter = nil
    end

    self.nicknameRenderNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.nicknameRenderNode#index"));
    self.nicknameRenderNodeOffset = Utils.getVectorNFromString(getXMLString(xmlFile, "vehicle.nicknameRenderNode#offset"), 3);
    if self.nicknameRenderNode == nil then
        if self.vehicleCharacter ~= nil and self.vehicleCharacter.characterDistanceRefNode ~= nil then
            self.nicknameRenderNode = self.vehicleCharacter.characterDistanceRefNode;
            if self.nicknameRenderNodeOffset == nil then
                self.nicknameRenderNodeOffset = {0,1.5,0};
            end
        else
            self.nicknameRenderNode = self.components[1].node;
        end
    end
    if self.nicknameRenderNodeOffset == nil then
        self.nicknameRenderNodeOffset = {0,4,0};
    end

    self.enterAnimation = getXMLString(xmlFile, "vehicle.enterAnimation#name");
end

---Called on deleting
-- @includeCode
function customSteerable:delete()
    g_currentMission:removeInteractiveVehicle(self);
    if self.vehicleCharacter ~= nil then
        self.vehicleCharacter:delete();
    end
    for _, camera in ipairs(self.cameras) do
        camera:delete();
    end
    if self.sampleToggleLights ~= nil then
        SoundUtil.deleteSample(self.sampleToggleLights);
    end
end

---Called on client side on join
-- @param integer streamId streamId
-- @param integer connection connection
-- @includeCode
function customSteerable:readStream(streamId, connection)
    local isControlled = streamReadBool(streamId);
    if isControlled then
        self.controllerName = streamReadString(streamId);
        local playerIndex = streamReadUIntN(streamId, PlayerUtil.sendNumBits)
        local playerColorIndex = streamReadUInt8(streamId)
        self:enterVehicle(false, playerIndex, playerColorIndex);
    end
end

---Called on server side on join
-- @param integer streamId streamId
-- @param integer connection connection
-- @includeCode
function customSteerable:writeStream(streamId, connection)
    if streamWriteBool(streamId, self.isControlled) then
        streamWriteString(streamId, self.controllerName);
        streamWriteUIntN(streamId, self.playerIndex, PlayerUtil.sendNumBits)
        streamWriteUInt8(streamId, self.playerColorIndex)
    end
end

function customSteerable:readUpdateStream(streamId, timestamp, connection)
end

function customSteerable:writeUpdateStream(streamId, connection, dirtyMask)
end

---Returns string with name of controller for game stats xml
-- @return string statesAttributes states attributes
-- @includeCode
function customSteerable:getXMLStatsAttributes()
    if self.isControlled and self.controllerName ~= nil then
        return string.format('controller="%s"', Utils.encodeToHTML(tostring(self.controllerName)));
    end
    return nil;
end

---Called on mouse event
-- @param float posX x position
-- @param float posY y position
-- @param boolean isDown is mouse button down
-- @param boolean isUp is mouse button up
-- @param integer button mouse button
-- @includeCode
function customSteerable:mouseEvent(posX, posY, isDown, isUp, button)
    if self.activeCamera ~= nil and self.isEntered then
        self.activeCamera:mouseEvent(posX, posY, isDown, isUp, button);
    end
end

function customSteerable:keyEvent(unicode, sym, modifier, isDown)
end

---Called on update
-- @param float dt time since last call in ms
-- @includeCode
function customSteerable:update(dt)
    if self:getIsActive() then

        if InputBinding.hasEvent(InputBinding.ENTER) and self.isEntered and not g_gui:getIsGuiVisible() then
            g_currentMission:onLeaveVehicle();
        end

        if InputBinding.hasEvent(InputBinding.RESET_HEAD_TRACKING) then
            centerHeadTracking();
        end

        if self.isClient then
            if self.steering ~= nil then
                local rotTime = self.rotatedTime;
                local baseRotation = self.steeringSpeed;

                if g_currentMission.controlledVehicle ~= self or not self.activeCamera.isInside then
                    baseRotation = Utils.getNoNil(self.steeringOutdoorRotation, self.steeringSpeed);
                else
                    baseRotation = Utils.getNoNil(self.steeringIndoorRotation, self.steeringSpeed);
                end

                if self.steeringOutdoorRotation ~= nil or self.steeringIndoorRotation ~= nil then
                    if self.rotatedTime > 0 then
                        rotTime = self.rotatedTime/math.abs(self.maxRotTime);
                    else
                        rotTime = self.rotatedTime/math.abs(self.minRotTime);
                    end
                end

                local rotation = baseRotation * rotTime;

                if self.steeringLastRotation ~= rotation then
                    self.steeringLastRotation = rotation;
                    setRotation(self.steering, 0, rotation*Utils.getNoNil(self.reverserDirection, 1), 0);

                    if self.vehicleCharacter ~= nil then
                        self.vehicleCharacter:setDirty();
                    end
                end
            end

            if self.isEntered and self.vehicleCharacter ~= nil then
                if self.vehicleCharacter.characterSpineNode ~= nil and self.vehicleCharacter.characterSpineSpeedDepended then
                    self.vehicleCharacter:setSpineDirty(self.lastSpeedAcceleration);
                end
            end
        end
    end

    if self.isEntered and self.isClient then
        if not g_currentMission.hasSpecialCamera then
            setCamera(self.activeCamera.cameraNode);
        end
        self.activeCamera:update(dt);

        -- do all the input handling
        local activeForInput = true;
        if g_gui:getIsGuiVisible() or g_currentMission.isPlayerFrozen then
            activeForInput = false;
        end

        if activeForInput then

            if InputBinding.hasEvent(InputBinding.CAMERA_SWITCH) then
                self:setActiveCameraIndex(self.camIndex + 1);
            end

            -- camera zoom is handled here, otherwise the controls are laggy or don't work at all
            if self.activeCamera.allowTranslation then
                if InputBinding.isPressed(InputBinding.CAMERA_ZOOM_IN) then
                    if InputBinding.getInputTypeOfDigitalAction(InputBinding.CAMERA_ZOOM_IN) == InputBinding.INPUTTYPE_MOUSE_WHEEL then
                        self.activeCamera:zoomSmoothly(-0.6);
                    else
                        self.activeCamera:zoomSmoothly(-0.005*dt);
                    end
                elseif InputBinding.isPressed(InputBinding.CAMERA_ZOOM_OUT) then
                    if InputBinding.getInputTypeOfDigitalAction(InputBinding.CAMERA_ZOOM_OUT) == InputBinding.INPUTTYPE_MOUSE_WHEEL then
                        self.activeCamera:zoomSmoothly(0.6);
                    else
                        self.activeCamera:zoomSmoothly(0.005*dt);
                    end
                end
            end

            if self:canToggleLight() then

                local playSound = false;

                if self.numLightTypes >= 1 and InputBinding.hasEvent(InputBinding.TOGGLE_LIGHT_FRONT) then
                    playSound = true;
                    local lightsTypesMask = bitXOR(self.lightsTypesMask, 2^0);
                    self:setLightsTypesMask(lightsTypesMask);
                elseif InputBinding.hasEvent(InputBinding.TOGGLE_LIGHTS) then
                    customSteerable.setNextLightsState(self);
                end

                if self.numLightTypes >= 2 then
                    -- work light back has light type 1
                    if InputBinding.hasEvent(InputBinding.TOGGLE_WORK_LIGHT_BACK) then
                        playSound = true;
                        local lightsTypesMask = bitXOR(self.lightsTypesMask, 2^1);
                        self:setLightsTypesMask(lightsTypesMask);
                    end

                    if self.numLightTypes >= 3 then
                        -- work light front has light type 2
                        if InputBinding.hasEvent(InputBinding.TOGGLE_WORK_LIGHT_FRONT) then
                            playSound = true;
                            local lightsTypesMask = bitXOR(self.lightsTypesMask, 2^2);
                            self:setLightsTypesMask(lightsTypesMask);
                        end

                        if self.numLightTypes >= 4 then
                            -- work light has light type 3
                            if InputBinding.hasEvent(InputBinding.TOGGLE_HIGH_BEAM_LIGHT) then
                                playSound = true;
                                local lightsTypesMask = bitXOR(self.lightsTypesMask, 2^3);
                                self:setLightsTypesMask(lightsTypesMask);
                            end
                        end
                    end
                end

                if playSound then
                    if self.sampleToggleLights ~= nil then
                        SoundUtil.playSample(self.sampleToggleLights, 1, 0, 1);
                    else
                        SoundUtil.playSample(g_currentMission.sampleToggleLights, 1, 0, 1);
                    end
                end

            end

        end

        -- update character visiblity
        if self.vehicleCharacter ~= nil then
            self.vehicleCharacter:updateVisibility();
        end
    end
end

function customSteerable:updateTick(dt)
end

---Called on draw
-- @includeCode
function customSteerable:draw()
    if not self.isEntered then
        return;
    end

    g_currentMission:addHelpButtonText(g_i18n:getText("action_exitVehicle"), InputBinding.ENTER, nil, GS_PRIO_NORMAL);

    if self:canToggleLight() and not self.isAttachable then
        if g_currentMission.environment.currentHour > 20 or  g_currentMission.environment.currentHour < 7 then
            g_currentMission:addHelpButtonText(g_i18n:getText("input_TOGGLE_LIGHTS"), InputBinding.TOGGLE_LIGHTS, nil, GS_PRIO_NORMAL);
        end
    end
end

---Called on ui info draw, renders nicknames in multiplayer
-- @includeCode
function customSteerable:drawUIInfo(superFunc)
    if superFunc ~= nil then
        superFunc(self);
    end

    if not self.isEntered and self.isClient and self:getIsActive() and self.isControlled and not g_gui:getIsGuiVisible() and g_currentMission.showHudEnv then
        local x,y,z = getWorldTranslation(self.nicknameRenderNode);
        local x1,y1,z1 = getWorldTranslation(getCamera())
        local distSq = Utils.vector3LengthSq(x-x1,y-y1,z-z1);
        if distSq <= 100*100 then
            x = x + self.nicknameRenderNodeOffset[1];
            y = y + self.nicknameRenderNodeOffset[2];
            z = z + self.nicknameRenderNodeOffset[3];

            Utils.renderTextAtWorldPosition(x,y,z, self.controllerName, getCorrectTextSize(0.02), 0)
        end
    end
end

---Enter vehicle
-- @param boolean isControlling is controlling vehicle
-- @param integer playerIndex index of player who enters the vehicle
-- @param integer playerColorIndex index of player color
-- @includeCode
function customSteerable:enterVehicle(isControlling, playerIndex, playerColorIndex)
    self.isControlled = true;
    if isControlling then
        self.isEntered = true;

        -- if head tracking is available we want to use the first indoor camera
        if g_gameSettings:getValue("isHeadTrackingEnabled") and isHeadTrackingAvailable() then
            for i,camera in pairs(self.cameras) do
                if camera.isInside then
                    self.camIndex = i;
                    break;
                end
            end
        end

        if g_gameSettings:getValue("resetCamera") then
            self.camIndex = 1
        end
        self:setActiveCameraIndex(self.camIndex);
    end

    if self.vehicleCharacter ~= nil and not self:getIsHired() then
        self.vehicleCharacter:loadCharacter(PlayerUtil.playerIndexToDesc[playerIndex].xmlFilename, playerColorIndex)
    end

    if self.enterAnimation ~= nil and self.playAnimation ~= nil then
        self:playAnimation(self.enterAnimation, 1, nil, true);
    end

    self.playerIndex = playerIndex
    self.playerColorIndex = playerColorIndex

    g_currentMission.controlledVehicles[self] = self;
    self:onEnter(isControlling)

    if self.isServer and not isControlling and g_currentMission.trafficSystem ~= nil and g_currentMission.trafficSystem.trafficSystemId ~= 0 then
        addTrafficSystemPlayer(g_currentMission.trafficSystem.trafficSystemId, self.components[1].node);
    end
end

---Leave vehicle
-- @includeCode
function customSteerable:leaveVehicle()
    if self.isServer and not self.isEntered and g_currentMission.trafficSystem ~= nil and g_currentMission.trafficSystem.trafficSystemId ~= 0 then
        removeTrafficSystemPlayer(g_currentMission.trafficSystem.trafficSystemId, self.components[1].node);
    end

    g_currentMission.interactionBlockTimer = 500
    self:onLeave()

    self.isControlled = false;
    if self.activeCamera ~= nil and self.isEntered then
        self.activeCamera:onDeactivate();
        SoundUtil.onCameraSwitched(false);
    end
    if self.vehicleCharacter ~= nil and self.disableCharacterOnLeave then
        self.vehicleCharacter:delete()
    end

    if self:getDeactivateOnLeave() then
        self:onDeactivate();
    else
        self:onDeactivateSounds();
    end

    if self.enterAnimation ~= nil and self.playAnimation ~= nil then
        self:playAnimation(self.enterAnimation, -1, nil, true);
    end

    self.currentLightState = 0

    self.isEntered = false;
    g_currentMission.controlledVehicles[self] = nil;
end

---Called on enter vehicle
-- @param boolean isControlling is player controlling the vehicle
-- @includeCode
function customSteerable:onEnter(isControlling)
    if self.forceSelectionOnEnter then
        local rootAttacherVehicle = self:getRootAttacherVehicle();
        rootAttacherVehicle:setSelectedImplement(rootAttacherVehicle:getImplementByObject(self));

    end;
end

function customSteerable:onLeave()
end

---Returns attributes and nodes to save
-- @param table nodeIdent node ident
-- @return string attributes attributes
-- @return string nodes nodes
-- @includeCode
function customSteerable:getSaveAttributesAndNodes(nodeIdent)
    local attributes = "";
    local nodes = "";
    return attributes, nodes;
end

---Toggle next light state
-- @includeCode
function customSteerable.setNextLightsState(self)
    if self.lightStates ~= nil and #self.lightStates > 0 then
        if self.sampleToggleLights ~= nil then
            SoundUtil.playSample(self.sampleToggleLights, 1, 0, 1);
        else
            SoundUtil.playSample(g_currentMission.sampleToggleLights, 1, 0, 1);
        end

        local currentLightState = self.currentLightState + 1;
        if currentLightState > #self.lightStates or (self.currentLightState == 0 and self.lightsTypesMask > 0) then
            currentLightState = 0;
        end

        local lightsTypesMask = 0;
        if currentLightState > 0 then
            for _, lightType in pairs(self.lightStates[currentLightState]) do
                lightsTypesMask = bitOR(lightsTypesMask, 2^lightType);
            end
        end

        self:setLightsTypesMask(lightsTypesMask);
        self.currentLightState = currentLightState;
    end
end

---Returns distance between given object and enterReferenceNode
-- @param integer object id of object
-- @return float distance distance
-- @includeCode
function customSteerable:getDistanceToObject(superFunc, object)
    local px, _, pz = getWorldTranslation(object);
    local vx, _, vz = getWorldTranslation(self.enterReferenceNode);
    local distance = Utils.vector2Length(px-vx, pz-vz);

    local superDistance = math.huge
    if superFunc ~= nil then
        superDistance = superFunc(self, object)
    end

    if distance < self.interactionRadius and distance < superDistance then
        self.interactionFlag = Vehicle.INTERACTION_FLAG_STEERABLE
        return distance
    end

    return superDistance;
end

---Returns interaction help text
-- @return string text text
-- @includeCode
function customSteerable:getInteractionHelp(superFunc)
    if self.interactionFlag == Vehicle.INTERACTION_FLAG_STEERABLE then
        return g_i18n:getText("action_enter");
    else
        if superFunc ~= nil then
            return superFunc(self)
        else
            return ""
        end
    end
end

---Interact
-- @includeCode
function customSteerable:interact(superFunc)
    if self.interactionFlag == Vehicle.INTERACTION_FLAG_STEERABLE then
        g_currentMission:requestToEnterVehicle(self)
    else
        if superFunc ~=  nil then
            return superFunc(self)
        end
    end
end

---Returns if vehicle is active for input
-- @param boolean onlyTrueIfSelected only true if selected
-- @param boolean activeIfIngameMessageShown active if ingame message is shown
-- @return boolean isActive is active for input
-- @includeCode
function customSteerable:getIsActiveForInput(superFunc, onlyTrueIfSelected, activeIfIngameMessageShown)
    if not self.isAttachable then
        return superFunc(self, onlyTrueIfSelected, activeIfIngameMessageShown);
    else
        if g_gui:getIsGuiVisible() or g_currentMission.isPlayerFrozen then
            return false;
        end
        if self.isEntered then
            if self.attacherVehicle ~= nil then
                return true;
            else
                return (not self.steerableNeedsAttacherVehicle) and (not self.isHired);
            end
        end
        return false;
    end
end

---Change active camera index
-- @param integer index index of camera to set
-- @includeCode
function customSteerable:setActiveCameraIndex(index)
    if self.activeCamera ~= nil then
        self.activeCamera:onDeactivate();
    end
    self.camIndex = index;
    if self.camIndex > self.numCameras then
        self.camIndex = 1;
    end
    self.activeCamera = self.cameras[self.camIndex];
    self.activeCamera:onActivate();
    self:onCameraChanged(self.activeCamera, self.camIndex)

    --
    SoundUtil.onCameraSwitched(self.activeCamera.isInside);
end

---Add cameras from tool
-- @param table cameras cameras to add
-- @includeCode
function customSteerable:addToolCameras(cameras)
    for _,toolCamera in pairs(cameras) do
        table.insert(self.cameras, toolCamera);
    end
    self.numCameras = #self.cameras;
end

---Remove cameras from tool
-- @param table cameras cameras to remove
-- @includeCode
function customSteerable:removeToolCameras(cameras)
    for _,toolCamera in pairs(cameras) do
        for j,camera in pairs(self.cameras) do
            if toolCamera == camera then
                table.remove(self.cameras, j);
                break;
            end
        end
    end
    self.numCameras = #self.cameras;
    if self.activeCamera ~= nil then
        self.activeCamera:onDeactivate();
    end
    if self.camIndex > self.numCameras then
        self.camIndex = 1;
    end
    local cameraId = getCamera();
    self:setActiveCameraIndex(self.camIndex);
    setCamera(cameraId);
end
